package pacman.state;


import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.event.KeyEvent;

import pacman.actors.Actor;
import pacman.actors.Ghost;
import pacman.actors.Player;
import pacman.ai.AIManager;
import pacman.game.Game;
import pacman.game.GameObject;
import pacman.game.Item;
import pacman.map.Map;
import pacman.util.Direction;

/**
 * StateGame is a mode of the program where the user can play the game of Pacman
 * various maps StateGame is a subclass of State
 * 
 * @author Ramsey Kant
 */
public class StateGame extends State {

    // Logic object references
    private Player player;
    private Map map;
    private AIManager ai;

    // Game vars
    private String mapName;
    private int currentLevel;
    private int sessionScore; // Overall score for the game session. The player
                              // object score is only the score for that life /
                              // level
    private int livesRemaining;
    private boolean gamePaused;
    private long pauseTime;

    // Map vars. Store them as class member vars to eliminate function call
    // overhead for getHeight/getWidth
    private int mapWidth;
    private int mapHeight;

    /**
     * StateGame Constructor
     * 
     * @param g
     *            Reference to the game supervisor
     */
    public StateGame(Game g) {
        super(g);
    }

    /**
     * Get the current session score. A session score is the total score of the
     * entire 'game' contained by the limited number of lives This is compounded
     * at game over and in win()
     * 
     * @see StateGame#win()
     * @see StateGame#lose()
     */
    public int getSessionScore() {
        return sessionScore;
    }

    // Public Methods

    /**
     * Reset the state of the game entirely (level 1)
     * 
     * @see State#reset()
     */
    @Override
    public void reset() {
        // Set game vars
        mapName = game.getStartMap();
        currentLevel = 0;
        sessionScore = 0;
        livesRemaining = 99;
        pauseTime = 0;

        // Respawn (start level 1)
        respawn(true);
    }

    /**
     * Respawns the player after a death or win. Similar to reset() except
     * running vars are saved
     * 
     * @param nextLevel
     *            Moves to the next level and corresponding map
     */
    public void respawn(boolean nextLevel) {
        gamePaused = true;
        pauseTime = System.currentTimeMillis() + 3000;

        // If we're jumping to the next level, reset everything
        if (nextLevel) {
            currentLevel = currentLevel + 1;

            // Force previous references out of scope
            player = null;
            map = null;
            ai = null;

            // Setup the game map
            game.getGraphicsContext().setBackground(Color.BLACK);
            map = new Map(mapName, 0.75);
            mapWidth = map.getWidth();
            mapHeight = map.getHeight();

            // Spawn the player
            player = map.getPlayer();

            // Setup AI
            ai = new AIManager(map, player, game.isDebugEnabled());

            // Slighly increase the game speed

        } else { // Player died, reset the map
            // Move all actors back to their spawn positions
            final int nActors = map.getNumActors();
            int i = 0;
			while (i < nActors) {
				final Actor a = map.getActor(i);
				if (a != null) {
					a.move(a.getSpawnX(), a.getSpawnY());
					a.setDead(false);
					if (a.getType() == GameObject.OBJECT_GHOST) {
						((Ghost) a).updatePath(null);
					}
				}
				i = i + 1;
			}
        }
    }

    /**
     * Main game logic for rendering and processing. Called by mainThreadLoop
     * 
     * @see Game#mainThreadLoop()
     * @see State#logic()
     */
    @Override
    public void logic() {
        if (map == null) {
            return;
        }

        final Graphics2D g = game.getGraphicsContext();

        // Offset the buffer so object's arent clipped by the window borders
        g.translate(10, 30);

        // Paint right UI with lives remaining, score, highscore etc
        g.setColor(Color.WHITE);
        g.setFont(new Font("Comic Sans MS", Font.BOLD, 24));
        g.drawString("PACMAN by Ramsey Kant", 680, 50);
        g.drawString("Score: " + player.getScore(), 750, 100);
        g.drawString("Total: " + sessionScore, 750, 150);
        g.drawString("Lives: " + livesRemaining, 750, 200);
        g.drawString("Level: " + currentLevel, 750, 250);

        // Execute game logic for all entites on the map
        if (!gamePaused) {
            ai.process();
            player.act();
        }

        // Check for player death. End the round if the player is dead
        if (player.isDead()) {
            lose();
            return;
        }

        // Check for a win (all dots collected)
        if (map.getDotsRemaining() <= 0) {
            win();
            return;
        }

        // Now run render logic
        // Paint boundaries and items
        Item item = null;
        int x = 0;
		while (x < mapWidth) {
			int y = 0;
			while (y < mapHeight) {
				final byte sType = map.getCollidable(x, y);
				g.setColor(Color.BLUE);
				switch (sType) {
				case 0:
					break;
				case GameObject.WALL_VERTICAL:
					g.fillRoundRect(x * map.CELL_SIZE + map.WALL_OVERLAP, y
							* map.CELL_SIZE, map.WALL_THICKNESS, map.CELL_SIZE,
							0, 0);
					break;
				case GameObject.WALL_HORIZONTAL:
					g.fillRoundRect(x * map.CELL_SIZE, y * map.CELL_SIZE
							+ map.WALL_OVERLAP, map.CELL_SIZE,
							map.WALL_THICKNESS, 0, 0);
					break;
				case GameObject.WALL_TOPLEFT:
					g.fillRoundRect(x * map.CELL_SIZE + (map.CELL_SIZE / 2), y
							* map.CELL_SIZE + map.WALL_OVERLAP,
							map.CELL_SIZE / 2, map.WALL_THICKNESS, 0, 0);
					g.fillRoundRect(x * map.CELL_SIZE + map.WALL_OVERLAP, y
							* map.CELL_SIZE + (map.CELL_SIZE / 2),
							map.WALL_THICKNESS, map.CELL_SIZE / 2, 0, 0);
					break;
				case GameObject.WALL_TOPRIGHT:
					g.fillRoundRect(x * map.CELL_SIZE, y * map.CELL_SIZE
							+ map.WALL_OVERLAP, map.CELL_SIZE / 2,
							map.WALL_THICKNESS, 0, 0);
					g.fillRoundRect(x * map.CELL_SIZE + map.WALL_OVERLAP, y
							* map.CELL_SIZE + (map.CELL_SIZE / 2),
							map.WALL_THICKNESS, map.CELL_SIZE / 2, 0, 0);
					break;
				case GameObject.WALL_BOTTOMLEFT:
					g.fillRoundRect(x * map.CELL_SIZE + (map.CELL_SIZE / 2), y
							* map.CELL_SIZE + map.WALL_OVERLAP,
							map.CELL_SIZE / 2, map.WALL_THICKNESS, 0, 0);
					g.fillRoundRect(x * map.CELL_SIZE + map.WALL_OVERLAP, y
							* map.CELL_SIZE, 12, map.CELL_SIZE / 2, 0, 0);
					break;
				case GameObject.WALL_BOTTOMRIGHT:
					g.fillRoundRect(x * map.CELL_SIZE, y * map.CELL_SIZE
							+ map.WALL_OVERLAP, map.CELL_SIZE / 2,
							map.WALL_THICKNESS, 0, 0);
					g.fillRoundRect(x * map.CELL_SIZE + map.WALL_OVERLAP, y
							* map.CELL_SIZE, map.WALL_THICKNESS,
							map.CELL_SIZE / 2, 0, 0);
					break;
				case GameObject.WALL_GHOSTBARRIER:
					g.setColor(Color.PINK);
					g.fillRoundRect(x * map.CELL_SIZE, y * map.CELL_SIZE
							+ map.WALL_OVERLAP, map.CELL_SIZE,
							map.WALL_THICKNESS / 2, 0, 0);
					break;
				default:
					break;
				}
				item = map.getItem(x, y);
				if (item != null) {
					item.paint(g);
				}
				y = y + 1;
			}
			x = x + 1;
		}

        // Paint actors ontop
        final int nActors = map.getNumActors();
        int i = 0;
		while (i < nActors) {
			final Actor a = map.getActor(i);
			if (a != null) {
				a.paint(g);
			}
			i = i + 1;
		}

        // Debug
        if (game.isDebugEnabled()) {
            g.setColor(Color.RED);
            g.drawString("DEBUG ON", 750, 650);
            /*
             * // Paint gridline overlay for(int i = 0; i < mapWidth; i++)
             * g.drawLine(i*map.CELL_SIZE, 0, i*map.CELL_SIZE,
             * mapHeight*map.CELL_SIZE); for(int i = 0; i < mapHeight; i++)
             * g.drawLine(0, i*map.CELL_SIZE, mapWidth*map.CELL_SIZE,
             * i*map.CELL_SIZE);
             */

            // Player X,Y coordinates bottom right
            g.drawString("positionX: " + player.getX(), 750, 675);
            g.drawString("positionY: " + player.getY(), 750, 700);
        }

        // Check for game pause and print pause status
        if (gamePaused) {
            g.setColor(Color.RED);
            g.setFont(new Font("Comic Sans MS", Font.BOLD, 24));
            g.drawString("PAUSED", 750, 500);
            if (pauseTime > System.currentTimeMillis()) {
                g.drawString(
                        "Pause ends in..." + ((pauseTime - System.currentTimeMillis()) / 1000),
                        750, 550);
            }
            if (pauseTime != 0 && System.currentTimeMillis() > pauseTime) {
                pauseTime = 0;
                gamePaused = false;
            }
            return;
        }
    }

    /**
     * Player has died or the Supervisor has decided to change the state
     * abruptly
     * 
     * @see State#end()
     */
    @Override
    public void end() {
        // Cleanup
        player = null;
        map = null;
    }

    /**
     * Player has won, move to the next level Called by logic()
     * 
     * @see StateGame#logic()
     */
    public void win() {
        sessionScore = sessionScore + (player.getScore());

        respawn(true);
    }

    /**
     * Player has died, reset() if lives remain. Otherwise, request a state
     * change thereby end()ing this state Called by logic()
     * 
     * @see StateGame#logic()
     */
    public void lose() {
        livesRemaining = livesRemaining - 1;

        if (livesRemaining > 0) {
            respawn(false);
        } else {
            if (currentLevel == 1) {
                sessionScore = player.getScore(); // win() never called, so
                                                  // score is the 1st level
                                                  // score
            }
            game.requestChangeState(State.STATE_SCOREBOARD);
        }
    }

    /**
     * Start automove in certain direction
     * 
     * @see java.awt.event.KeyListener#keyPressed(java.awt.event.KeyEvent)
     */
    @Override
    public void keyPressed(KeyEvent e) {
        if (player == null) {
            return;
        }

        switch (e.getKeyCode()) {
            case KeyEvent.VK_UP:
                player.setMoveDirection(Direction.up);
                break;
            case KeyEvent.VK_RIGHT:
                player.setMoveDirection(Direction.right);
                break;
            case KeyEvent.VK_DOWN:
                player.setMoveDirection(Direction.down);
                break;
            case KeyEvent.VK_LEFT:
                player.setMoveDirection(Direction.left);
                break;
            case KeyEvent.VK_SPACE:
                player.setMoveDirection(Direction.stop);
                break;
            case KeyEvent.VK_P:
                // Don't interupt system pauses
                if (pauseTime < System.currentTimeMillis()) {
                    gamePaused = !gamePaused;
                }
                break;
            case KeyEvent.VK_V:
                game.toggleDebug();
                // AI debug
                ai.setDebugEnabled(game.isDebugEnabled());
                break;
            case KeyEvent.VK_0:
                // game.changeState(STATE_MENU);
                break;
            default:
                break;
        }
    }
}
