package pacman.actors;

import java.awt.Color;
import java.awt.Graphics2D;

import pacman.ai.AIManager;
import pacman.game.GameObject;
import pacman.map.Map;
import pacman.state.StateGame;
import pacman.util.Direction;
import pacman.util.RequestedDirectionBuffer;

/**
 * An actor is any object that has a degree of autonomy or intelligent input
 * (Human / AIManager) dictating the object's behavior ingame Subclass of
 * GameObject
 * 
 * @author Ramsey Kant
 */
public abstract class Actor extends GameObject {

    /** Whether the actor is alive or not. */
    protected boolean f012;

    /** The x-position on the map where the actor gets spawned */
    protected int f112;
    
    /** The y-position on the map where the actor gets spawned */
    protected int f212;

    /** The direction in which the actor is currently oriented. */
    protected Direction f312;
    
    /** The actor's direction requested by user input. Uses a buffer to
     *  remember requests over some steps to simplify timing for the user.
     *  Computer-controlled actors (ghosts) ignore this. */
    protected RequestedDirectionBuffer f412;
    
    /** The size of the direction request buffer. */
    private final static int f512 = 6;

    /** The current orientation angle of the actor. Ignored by actors that do
     *  not have to orient (aka ghosts). */
    protected int f612;
    
    /** The x-position delta to the current map cell, caused by movement
     *  in pixels. */
    protected float f712;
    
    /** The y-position delta to the current map cell, caused by movement
     *  in pixels. */
    protected float f812;
    
    /** The movement speed of the actor */
    protected float f912;

    /**
     * Actor Class Constructor
     * 
     * @param v0
     *            Object type that is an actor
     * @param v1
     *            Base color of the actor
     * @param v2
     *            Reference to the global map object
     * @param v3
     *            X coordinate to spawn the actor at
     * @param v4
     *            Y coordinate to spawn the actor at
     * @see GameObject
     */
    public Actor(int v0, Color v1, Map v2, int v3, int v4) {
        super(v0, v1, v2, v3, v4);

        f012 = false;

        // Movement
        f112 = v3;
        f212 = v4;
        f312 = Direction.f41;
        f412 = new RequestedDirectionBuffer(f512);
        f612 = 0;
        f712 = 0;
        f812 = 0;
        f912 = (float) (5d * f1916.f55);
    }

    // Getters and Setters

    /**
     * Returns the original X coordinate the Actor was given in the constructor
     * 
     * @return the X coordinate of the spawn point
     */
    public int m012() {
        return f112;
    }

    /**
     * Returns the original Y coordinate the Actor was given in the constructor
     * 
     * @return the Y coordinate of the spawn point
     */
    public int m112() {
        return f212;
    }

    /**
     * Set the death status of the actor. Used by StateGame and AIManager to
     * determine if the player / ghost has died
     */
    public void m212(boolean v5) {
        f012 = v5;
    }

    /**
     * Get dead status
     * 
     * @return True if dead, false if alive
     * @see Actor#m212(boolean)
     */
    public boolean m312() {
        return f012;
    }

    /**
     * Speed is the number of pixels an actor moves across the screen in a given
     * cycle. A full position change is the number of pixels defined in
     * Map.CELL_SIZE
     * 
     * @param v6
     *            New Speed
     */
    public void m412(float v6) {
        f912 = v6;
    }

    /**
     * Get the current speed of the actor
     * 
     * @return Current speed
     * @see Actor#m412(float)
     */
    public float m512() {
        return f912;
    }

    /**
     * Set the direction actor should travel in. Player uses this to determine
     * the direction to "auto-move" to Ghosts ignore what is set by this
     * function because their direction is determined within act() based on the
     * path
     */
    public void m612(Direction v7) {
        f412.m114(v7);
    }

    // Public Methods

    /**
     * Attempt to move the actor to the given x,y location. This method will
     * check if a coordinate is valid with the Map class method canMove(). It is
     * not necessary to call canMove() before this function
     * 
     * @param v8
     *            A x coordinate to move to
     * @param v9
     *            A y coordinate to move to
     * @return True if the move succeeded. False if otherwise
     * @see Map#m195(Actor, int, int)
     */
    public boolean m712(int v8, int v9) {
        final boolean v10;
		v10 = f1916.m195(this, v8, v9);
        if (v10) {
            f1716 = v8;
            f1816 = v9;
        }
        return v10;
    }

    /**
     * The primary logic function for actors. StateGame calls this for players
     * directly in logic() and the AIManager calls this for ghosts in process()
     * 
     * @see GameObject#m516()
     * @see StateGame#m28()
     * @see AIManager#m20()
     */
    @Override
    public abstract void m516();

    /**
     * 
     * @see GameObject#m616(java.awt.Graphics2D)
     */
    @Override
    public abstract void m616(Graphics2D v11);

}
