package pacman.map;

import java.awt.Color;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;

import pacman.actors.Actor;
import pacman.actors.Ghost;
import pacman.actors.Player;
import pacman.game.GameObject;
import pacman.game.Item;
import pacman.util.Direction;

/**
 * Map class keeps track objects on the playing grid, helper methods to make
 * movement decisions, and export/import methods for the editor
 * 
 * @author Ramsey Kant
 */
public class Map {
    // Map parameters (width & height represent # of cells)
    private int f05;
    private int f15;
    public final int f25;
    public final int f35;
    public final int f45;
    public final double f55;

    // Instance vars
    private byte f65[][];
    private Item f75[][];
    private ArrayList<Actor> f85;
    private int f95;

    /**
     * Class constructor, inits a blank map based on a width, height, and cell
     * size Used in the editor
     * 
     * @param v0
     *            Width of the map
     * @param v1
     *            Height of the map
     * @param cs
     *            Size of individual cells in pixels
     */
    public Map(int v0, int v1, double v2) {
        // Set map parameters
        f05 = v0;
        f15 = v1;
        f55 = v2;
        f25 = (int) (32 * v2);
        f35 = (int) (12 * v2);
        f45 = (int) (10 * v2);
        f95 = 0;

        // Initialize collideMap, a 2D array that contains all static collidable
        // GameObjects
        // We use this for fast lookup during collision detection and AI
        // movement paths
        f65 = new byte[f05][f15];

        // Initialize itemMap, a 2D array that contains items (dots, powerups,
        // cherry) on the map
        f75 = new Item[f05][f15];

        // Create m_objects, an arraylist with all actorList
        f85 = new ArrayList<Actor>();
    }

    /**
     * Class Constructor that reads the map data from filename
     * 
     * @param v3
     *            The file name of the map to read contents from
     * @param cs
     *            Size of individual cells in pixels. This is something that
     *            should be deteremined by graphics, not the mapfile
     */
    public Map(String v3, double v4) {
        // Set the cell size
        f55 = v4;
        f25 = (int) (32 * v4);
        f35 = (int) (12 * v4);
        f45 = (int) (10 * v4);

        // Read contents of the map file
        m245(v3);
    }

    /**
     * The width of the map originally set in the constructor
     * 
     * @return The width of the map
     */
    public int m05() {
        return f05;
    }

    /**
     * The height of the map originally set in the constructor
     * 
     * @return The height of the map
     */
    public int m15() {
        return f15;
    }

    /**
     * Get the number of actorList on the map (the size of the actorList
     * ArrayList)
     * 
     * @return Number of actorList
     */
    public int m25() {
        return f85.size();
    }

    /**
     * Return the collidable map (a 2d array of bytes which correspond to the
     * collidable types defined in GameObject)
     * 
     * @return collidable map (collideMap)
     */
    public byte[][] m35() {
        return f65;
    }

    /**
     * Return the item map (a 2D array of Item objects)
     * 
     * @return item map (itemMap)
     */
    public Item[][] m45() {
        return f75;
    }

    /**
     * Return the number of dots remaining on the map. This is tracked by the
     * dotsRemaining local var (not a loop and count in itemMap)
     * 
     * @return dots remaining
     */
    public int m55() {
        return f95;
    }

    /**
     * Add a collidable (by type) to the collideMap
     * 
     * @param v5
     *            X coordinate
     * @param v6
     *            Y coordinate
     * @param v7
     *            Type of collidable
     * @return True if successful
     */
    public boolean m65(int v5, int v6, byte v7) {
        // Check bounds
        if (v5 < 0 || v6 < 0 || v5 >= f05 || v6 >= f15) {
            return false;
        }

        // Check if theres already something there
        if (f65[v5][v6] > 0) {
            return false;
        }

        // Add to the collideMap
        f65[v5][v6] = v7;
        return true;
    }

    /**
     * Put a new item to the item map
     * 
     * @param v8
     *            Item
     * @return True if successful
     */
    public boolean m75(Item v8) {
        if (v8 == null) {
            return false;
        }

        final int v9;
		v9 = v8.m316();
        final int v10;
		v10 = v8.m416();
        if (v9 < 0 || v10 < 0 || v9 >= f05 || v10 >= f15) {
            return false;
        }

        // Add to the itemMap
        if (v8.m016() == GameObject.f016) {
            f95++;
        }
        f75[v9][v10] = v8;
        return true;
    }

    /**
     * Put a new actor in the map (actorList ArrayList)
     * 
     * @param v11
     *            Actor
     * @return True if successful
     */
    public boolean m85(Actor v11) {
        if (v11 == null) {
            return false;
        }

        final int v12;
		v12 = v11.m316();
        final int v13;
		v13 = v11.m416();
        if (v12 < 0 || v13 < 0 || v12 >= f05 || v13 >= f15) {
            return false;
        }

        // Add to the array list
        f85.add(v11);
        return true;
    }

    /**
     * Return a value at (x,y) in the collision map
     * 
     * @param v14
     *            X Coordinate
     * @param v15
     *            Y Coordinate
     * @return Integer that represents the collision object
     */
    public byte m95(int v14, int v15) {
        // Check bounds
        if (v14 < 0 || v15 < 0 || v14 >= f05 || v15 >= f15) {
            return -1;
        }

        return f65[v14][v15];
    }

    /**
     * Return an item at coordinate (x,y) from within the item map (itemMap)
     * 
     * @param v16
     *            X Coordinate
     * @param v17
     *            Y Coordinate
     * @return Item the item that is found at (x,y)
     */
    public Item m105(int v16, int v17) {
        // Check bounds
        if (v16 < 0 || v17 < 0 || v16 >= f05 || v17 >= f15) {
            return null;
        }

        return f75[v16][v17];
    }

    /**
     * Return an actor at index in the actorList ArrayList
     * 
     * @param v18
     *            Index in actorList
     * @return Actor (null if non-existant)
     */
    public Actor m115(int v18) {
        Actor v19;
		v19 = null;
        try {
            v19 = f85.get(v18);
        } catch (final IndexOutOfBoundsException v20) {
            v20.printStackTrace();
        }
        return v19;
    }

    /**
     * Find and return the player object within the local actorList ArrayList
     * 
     * @return The player object. null if not found
     */
    public Player m125() {
        // Get from the object map
        for (final Actor v21 : f85) {
            if (v21.m016() == GameObject.f316) {
                return (Player) v21;
            }
        }

        return null;
    }

    /**
     * Return an actor at coordinate (x,y)
     * 
     * @param v22
     *            X Coordinate
     * @param v23
     *            Y Coordinate
     * @param v24
     *            If true, ignore a "Player" actor at (x,y)
     * @return Actor (null if an actor doesn't exist at the position)
     */
    public Actor m135(int v22, int v23, boolean v24) {
        // Check bounds
        if (v22 < 0 || v23 < 0 || v22 >= f05 || v23 >= f15) {
            return null;
        }

        // Get from the object map
        for (final Actor v25 : f85) {
            if (v24 && v25.m016() == GameObject.f316) {
                continue;
            }

            if (v25.m316() == v22 && v25.m416() == v23) {
                return v25;
            }
        }

        return null;
    }

    /**
     * Remove an actor from actorList based on index. Be careful when using
     * this! Just because an actor isn't in the map doesn't mean it's not
     * 'alive' This is primarily for the editor
     * 
     * @param v26
     *            Index of the actor
     */
    public void m145(int v26) {
        f85.remove(v26);
    }

    /**
     * Remove an item from the item array by coordinate (x, y)
     * 
     * @param v27
     *            X coordinate of the item
     * @param v28
     *            Y coordinate of the item
     */
    public void m155(int v27, int v28) {
        // Check bounds
        if (v27 < 0 || v28 < 0 || v27 >= f05 || v28 >= f15) {
            return;
        }

        if (f75[v27][v28].m016() == GameObject.f016) {
            f95--;
        }

        f75[v27][v28] = null;
    }

    /**
     * Remove everything at coordiante (x,y) Used by the editor only
     * 
     * @param v29
     *            X coordinate
     * @param v30
     *            Y coordinate
     * @return boolean True if something was removed, false if otherwise
     */
    public boolean m165(int v29, int v30) {
        boolean v31;
		v31 = false;

        // Check bounds
        if (v29 < 0 || v30 < 0 || v29 >= f05 || v30 >= f15) {
            return false;
        }

        // Remove any collidable
        if (f65[v29][v30] != 0) {
            f65[v29][v30] = 0;
            v31 = true;
        }

        // Remove any item
        if (f75[v29][v30] != null) {
            f75[v29][v30] = null;
            v31 = true;
        }

        int v32;
		// Remove any actor
        for (v32 = 0; v32 < f85.size(); v32++) {
            Actor v33;
			v33 = f85.get(v32);
            if (v33.m316() == v29 && v33.m416() == v30) {
                f85.remove(v32);
                v33 = null;
                v32--;
                v31 = true;
            }
        }

        return v31;
    }

    /**
     * Find the distance (Manhattan) between two objects
     * 
     * @param v34
     *            GameObject at the initial position
     * @param v35
     *            GameObject at the end position
     * @return Distance (integer)
     */
    public int m175(GameObject v34, GameObject v35) {
        return (int) Math.sqrt(Math.pow(Math.abs(v34.m316() - v35.m316()), 2)
                + Math.pow(Math.abs(v34.m416() - v35.m416()), 2));
    }

    /**
     * Check if a coordinate is completely empty (void of actorList, items, and
     * collissions) Used by the editor
     * 
     * @param v36
     *            A x coordinate to move to
     * @param v37
     *            A y coordinate to move to
     * @return True if empty. False if otherwise
     */
    public boolean m185(int v36, int v37) {
        // Check bounds
        if (v36 < 0 || v37 < 0 || v36 >= f05 || v37 >= f15) {
            return false;
        }

        // Check if the Object is hitting something on the collideMap
        if (m95(v36, v37) != 0) {
            return false;
        }

        // Check if object is hitting something on the itemMap
        if (m105(v36, v37) != null) {
            return false;
        }

        // Actor collission
        if (m135(v36, v37, false) != null) {
            return false;
        }

        return true;
    }

    /**
     * Move attempt method. Changes the position the map of the game object if
     * there are no obstructions
     * 
     * @param v38
     *            The actor object trying to move
     * @param v39
     *            A x coordinate to move to
     * @param v40
     *            A y coordinate to move to
     * @return True if the move succeeded. False if otherwise
     */
    public boolean m195(Actor v38, int v39, int v40) {
        if (v38 == null) {
            return false;
        }

        // Check bounds
        if (!m215(v39, v40)) {
            return false;
        }

        // Check if the Object is hitting something on the collideMap
        if (m95(v39, v40) != 0) {
            return false;
        }

        // Allow the Actor to move
        return true;
    }

    public boolean m205(Actor v41, Direction v42) {
        int v43;
		v43 = v41.m316();
        int v44;
		v44 = v41.m416();
        switch (v42) {
            case f01:
                v44--;
                break;
            case f11:
                v43++;
                break;
            case f21:
                v44++;
                break;
            case f31:
                v43--;
                break;
            case f41:
                return true;
        }

        return m195(v41, v43, v44);
    }

    private boolean m215(int v45, int v46) {
        return v45 > 0 && v46 > 0 && v45 <= f05 && v46 <= f15;
    }

    /**
     * Get the cost of moving through the given tile. This can be used to make
     * certain areas more desirable. A simple and valid implementation of this
     * method would be to return 1 in all cases.
     * 
     * @param v47
     *            The mover that is trying to move across the tile
     * @param v48
     *            The x coordinate of the tile we're moving from
     * @param v49
     *            The y coordinate of the tile we're moving from
     * @param v50
     *            The x coordinate of the tile we're moving to
     * @param v51
     *            The y coordinate of the tile we're moving to
     * @return The relative cost of moving across the given tile
     */
    public float m225(Actor v47, int v48, int v49, int v50, int v51) {
        return 1;
    }

    /**
     * Write the contents of this map to a file in the correct format
     * 
     * @param v52
     *            File name of the map
     */
    public void m235(String v52) {
        FileOutputStream v53;
        DataOutputStream v54;

        try {
            v53 = new FileOutputStream(v52);
            v54 = new DataOutputStream(v53);

            // Write the map file magic
            v54.writeUTF("RKPACMAP");

            // Write map width & height
            v54.writeInt(f05);
            v54.writeInt(f15);

            int v55;
			// Write the collision map
            for (v55 = 0; v55 < f05; v55++) {
                int v56;
				for (v56 = 0; v56 < f15; v56++) {
                    v54.write(f65[v55][v56]);
                }
            }

            Item v57;
			v57 = null;
            int v58;
			for (v58 = 0; v58 < f05; v58++) {
                int v59;
				for (v59 = 0; v59 < f15; v59++) {
                    v57 = f75[v58][v59];

                    // If an item doesnt exist at (x,y), write 'false' for
                    // nonexistant and continue
                    if (v57 == null) {
                        v54.writeBoolean(false);
                        continue;
                    }
                    v54.writeBoolean(true);

                    // Write properties of the item
                    v54.writeInt(v57.m016());
                    v54.writeInt(v57.m316());
                    v54.writeInt(v57.m416());
                    v54.writeInt(v57.m116().getRGB());
                    if (v57.m016() == GameObject.f716) {
                        v54.writeInt(v57.m17());
                        v54.writeInt(v57.m27());
                    }
                }
            }

            // Write the number of actorList, then all actor data
            v54.writeInt(f85.size());
            for (final Actor v60 : f85) {
                v54.writeInt(v60.m016());
                v54.writeInt(v60.m316());
                v54.writeInt(v60.m416());
                v54.writeInt(v60.m116().getRGB());
                if (v60.m016() == GameObject.f416) {
                    v54.writeBoolean(((Ghost) v60).m218());
                }
            }

            v54.close();
            v53.close();
        } catch (final IOException v61) {
            System.out.println("Failed to write map file: " + v61.getMessage());
        }
    }

    /**
     * Read a file with map contents and set the properties in this map Called
     * by the constructor.
     * 
     * @param v62
     *            File name of the map
     */
    private void m245(String v62) {

        FileInputStream v63;
        DataInputStream v64;

        try {
            v63 = new FileInputStream(v62);
            v64 = new DataInputStream(v63);

            // Check for the magic
            if (!v64.readUTF().equals("RKPACMAP")) {
                System.out.println("Not a map file!");
                return;
            }

            // Read map width & height
            f05 = v64.readInt();
            f15 = v64.readInt();
            f95 = 0;

            // Initialize collideMap, a 2D array that contains all static
            // collidable GameObjects
            // We use this for fast lookup during collision detection and AI
            // movement paths
            f65 = new byte[f05][f15];

            // Initialize itemMap, a 2D array that contains items (dots,
            // powerups, cherry) on the map
            f75 = new Item[f05][f15];

            // Create m_objects, an arraylist with all actorList
            f85 = new ArrayList<Actor>();

            int v65;
			// Read the collision map
            for (v65 = 0; v65 < f05; v65++) {
                int v66;
				for (v66 = 0; v66 < f15; v66++) {
                    m65(v65, v66, v64.readByte());
                }
            }

            int v67;
			// Read the item map
            for (v67 = 0; v67 < f05; v67++) {
                int v68;
				for (v68 = 0; v68 < f15; v68++) {
                    // If an item doesnt exist at (x,y), continue
                    if (!v64.readBoolean()) {
                        continue;
                    }

                    final int v69;
					v69 = v64.readInt();
                    final int v70;
					v70 = v64.readInt();
                    final int v71;
					v71 = v64.readInt();
                    final Color v72;
					v72 = new Color(v64.readInt());
                    m75(new Item(v69, v72, this, v70, v71));
                    if (v69 == GameObject.f716) {
                        final int v73;
						v73 = v64.readInt();
                        final int v74;
						v74 = v64.readInt();
                        f75[v70][v71].m07(v73, v74);
                    }
                }
            }

            final int v75;
			v75 = v64.readInt();
            int v76;
			for (v76 = 0; v76 < v75; v76++) {
                final int v77;
				v77 = v64.readInt();
                final int v78;
				v78 = v64.readInt();
                final int v79;
				v79 = v64.readInt();
                final Color v80;
				v80 = new Color(v64.readInt());
                if (v77 == GameObject.f316) {
                    m85(new Player(this, v78, v79));
                } else if (v77 == GameObject.f416) {
                    final boolean v81;
					v81 = v64.readBoolean();
                    m85(new Ghost(v80, this, v78, v79, v81));
                } /*
                   * else { addActor(new Actor(t, c, this, ix, iy)); }
                   */
            }

            v64.close();
            v63.close();
        } catch (final IOException v82) {
            System.out.println("Failed to read map file: " + v82.getMessage());
        }
    }

}
