package pacman.actors;

import java.awt.Color;
import java.awt.Graphics2D;

import pacman.ai.AIManager;
import pacman.game.GameObject;
import pacman.map.Map;
import pacman.state.StateGame;
import pacman.util.Direction;
import pacman.util.RequestedDirectionBuffer;

/**
 * An actor is any object that has a degree of autonomy or intelligent input
 * (Human / AIManager) dictating the object's behavior ingame Subclass of
 * GameObject
 * 
 * @author Ramsey Kant
 */
public abstract class Actor extends GameObject {

    /** Whether the actor is alive or not. */
    protected boolean f09;

    /** The x-position on the map where the actor gets spawned */
    protected int f19;
    
    /** The y-position on the map where the actor gets spawned */
    protected int f29;

    /** The direction in which the actor is currently oriented. */
    protected Direction f39;
    
    /** The actor's direction requested by user input. Uses a buffer to
     *  remember requests over some steps to simplify timing for the user.
     *  Computer-controlled actors (ghosts) ignore this. */
    protected RequestedDirectionBuffer f49;
    
    /** The size of the direction request buffer. */
    private final static int f59 = 6;

    /** The current orientation angle of the actor. Ignored by actors that do
     *  not have to orient (aka ghosts). */
    protected int f69;
    
    /** The x-position delta to the current map cell, caused by movement
     *  in pixels. */
    protected float f79;
    
    /** The y-position delta to the current map cell, caused by movement
     *  in pixels. */
    protected float f89;
    
    /** The movement speed of the actor */
    protected float f99;

    /**
     * Actor Class Constructor
     * 
     * @param v0
     *            Object type that is an actor
     * @param v1
     *            Base color of the actor
     * @param v2
     *            Reference to the global map object
     * @param v3
     *            X coordinate to spawn the actor at
     * @param v4
     *            Y coordinate to spawn the actor at
     * @see GameObject
     */
    public Actor(int v0, Color v1, Map v2, int v3, int v4) {
        super(v0, v1, v2, v3, v4);

        f09 = false;

        // Movement
        f19 = v3;
        f29 = v4;
        f39 = Direction.f40;
        f49 = new RequestedDirectionBuffer(f59);
        f69 = 0;
        f79 = 0;
        f89 = 0;
        f99 = (float) (5d * f1910.f55);
    }

    // Getters and Setters

    /**
     * Returns the original X coordinate the Actor was given in the constructor
     * 
     * @return the X coordinate of the spawn point
     */
    public int m09() {
        return f19;
    }

    /**
     * Returns the original Y coordinate the Actor was given in the constructor
     * 
     * @return the Y coordinate of the spawn point
     */
    public int m19() {
        return f29;
    }

    /**
     * Set the death status of the actor. Used by StateGame and AIManager to
     * determine if the player / ghost has died
     */
    public void m29(boolean v5) {
        f09 = v5;
    }

    /**
     * Get dead status
     * 
     * @return True if dead, false if alive
     * @see Actor#m29(boolean)
     */
    public boolean m39() {
        return f09;
    }

    /**
     * Speed is the number of pixels an actor moves across the screen in a given
     * cycle. A full position change is the number of pixels defined in
     * Map.CELL_SIZE
     * 
     * @param v6
     *            New Speed
     */
    public void m49(float v6) {
        f99 = v6;
    }

    /**
     * Get the current speed of the actor
     * 
     * @return Current speed
     * @see Actor#m49(float)
     */
    public float m59() {
        return f99;
    }

    /**
     * Set the direction actor should travel in. Player uses this to determine
     * the direction to "auto-move" to Ghosts ignore what is set by this
     * function because their direction is determined within act() based on the
     * path
     */
    public void m69(Direction v7) {
        f49.m111(v7);
    }

    // Public Methods

    /**
     * Attempt to move the actor to the given x,y location. This method will
     * check if a coordinate is valid with the Map class method canMove(). It is
     * not necessary to call canMove() before this function
     * 
     * @param v8
     *            A x coordinate to move to
     * @param v9
     *            A y coordinate to move to
     * @return True if the move succeeded. False if otherwise
     * @see Map#m195(Actor, int, int)
     */
    public boolean m79(int v8, int v9) {
        final boolean v10;
		v10 = f1910.m195(this, v8, v9);
        if (v10) {
            f1710 = v8;
            f1810 = v9;
        }
        return v10;
    }

    /**
     * The primary logic function for actors. StateGame calls this for players
     * directly in logic() and the AIManager calls this for ghosts in process()
     * 
     * @see GameObject#m510()
     * @see StateGame#m220()
     * @see AIManager#m27()
     */
    @Override
    public abstract void m510();

    /**
     * 
     * @see GameObject#m610(java.awt.Graphics2D)
     */
    @Override
    public abstract void m610(Graphics2D v11);

}
