package pacman.game;

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Frame;
import java.awt.Graphics2D;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.image.BufferStrategy;

import pacman.state.State;
import pacman.state.StateEditor;
import pacman.state.StateGame;
import pacman.state.StateMenu;
import pacman.state.StateScoreboard;

/**
 * The Game Supervisor. This class implements that core program logic, state
 * management, and graphics.
 * 
 * @author Ramsey Kant
 */
public class Game extends Canvas {
    private static final long f020 = 1L;

    // Debug vars
    private boolean f120;

    // Threading
    private boolean f220;

    // Graphics variables
    private Frame f320;
    public final int f420;
    public final int f520;
    private BufferStrategy f620;

    // State
    private int f720;
    private State f820;
    private boolean f920;
    private int f1020;
    private String f1120;

    /**
     * Class Constructor Set's up graphics and put's game logic into a startup
     * state by calling init()
     * 
     * @param v0
     *            Resolution X
     * @param v1
     *            Resolution Y
     * @see Game#m020()
     */
    public Game(int v0, int v1) {
        // Set resolution settings
        f420 = v0;
        f520 = v1;

        // Init game
        m020();
    }

    /**
     * Startup functionality for the program called by the constructor
     */
    private void m020() {
        // Debug vars
        f120 = false;

        f1120 = "test.map";
        f920 = false;

        // Setup the game frame
        f320 = new Frame("Pacman");
        f320.setLayout(null);
        setBounds(0, 0, f420, f520);
        f320.add(this);
        f320.setSize(f420, f520);
        f320.setResizable(false);
        f320.setVisible(true);

        // Set the exit handler with an anonymous class
        f320.addWindowListener(new WindowAdapter() {
            @Override
            public void windowClosing(WindowEvent v2) {
                // Exit main thread
                f220 = false;
            }
        });

        // Setup double buffering
        setIgnoreRepaint(true); // We'll handle repainting
        createBufferStrategy(2);
        f620 = getBufferStrategy();

        f220 = true;
    }

    // Getter and Setter methods

    /**
     * Get the Frame object encapsulating the program
     * 
     * @return The frame
     */
    public Frame m120() {
        return f320;
    }

    /**
     * Get a 'handle' of the current graphics buffer for drawing
     * 
     * @return The Graphics2D buffer
     */
    public Graphics2D m220() {
        return (Graphics2D) f620.getDrawGraphics();
    }

    /**
     * Get the name of the map to be loaded in StateGame
     * 
     * @return Map name (with .map extension)
     */
    public String m320() {
        return f1120;
    }

    /**
     * Set the default starting map (set by menu)
     * 
     * @param v3
     *            The name of the map to load (with the .map extension)
     */
    public void m420(String v3) {
        f1120 = v3;
    }

    /**
     * Return the current debug setting
     * 
     * @return True if debug setting is on
     * @see Game#m620()
     */
    public boolean m520() {
        return f120;
    }

    /**
     * Toggle debugging. Facilities like AIManager use this flag to display
     * diagnostic information like AI paths
     */
    public void m620() {
        f120 = !f120;
    }

    // Public Methods

    /**
     * Called by other states to safely change currentState. This is done so the
     * currentState's logic can finish
     * 
     * @see Game#m820()
     */
    public void m720(int v4) {
        f1020 = v4;
        f920 = true;
    }

    /**
     * The main game loop that handles graphics and game state determination
     */
    public void m820() {
    	final long v5;
		v5 = 20;
    	long v6;
		v6 = 0;
    	
        while (f220) {
            
            if ((v6 + v5) > System.currentTimeMillis()){
                continue;
            }
            
            v6 = System.currentTimeMillis();
        	
            // If a state change was requested, execute it now
            if (f920) {
                f920 = false;
                m920(f1020);
                continue;
            }

            final Graphics2D v7;
			v7 = m220();

            // Wipe the screen
            v7.setColor(Color.black);
            v7.fillRect(0, 0, f420, f520);

            // Run the logic of the current game state here
            f820.m28();
            
            // Show the new buffer
            v7.dispose();
            f620.show();
            
            try {
                Thread.sleep(10);
            } catch (InterruptedException v8) {
                // TODO Auto-generated catch block
                v8.printStackTrace();
            }

        }
    }

    // Private Methods

    /**
     * Change the state of the game. Called in mainThreadLogic()
     * 
     * @param v9
     *            The state to set. Must match the static vars above
     * @see Game#m720(int)
     * @see Game#m820()
     */
    private void m920(int v9) {
        // Cleanup for the outgoing state
        if (f820 != null) {
            f320.removeKeyListener(f820);
            removeKeyListener(f820);
            f820.m38();
        }

        // Set the new state type
        f720 = v9;

        // Instance the new state (reset() is called in the construtor)
        switch (f720) {
            case State.f28:
                f820 = new StateGame(this);
                break;
            case State.f18:
                f820 = new StateScoreboard(this);
                /*
                 * StateGame sb = new StateScoreboard(); int newScore = 0;
                 * 
                 * // If the previous state was STATE_GAME, pull the session
                 * score and pass it to the scoreboard if(currentState
                 * instanceof StateGame)
                 * sb.addScore((int)((StateGame)currentState
                 * ).getSessionScore()));
                 * 
                 * currentState = sb;
                 */
                break;
            case State.f48:
                f820 = new StateEditor(this);
                break;
            case State.f08:
                f820 = new StateMenu(this);
                break;
            case State.f58:
                f820 = null;
                f220 = false;
                break;
            default:
                break;
        }

        // Setup input handler and reset()
        if (f820 != null) {
            f320.addKeyListener(f820);
            addKeyListener(f820);
        }
    }
}
