package pacman.map;
import java.util.ArrayList;
import java.util.Collections;

import pacman.actors.Actor;
import pacman.ai.AStarHeuristic;

/**
 * A path finder implementation that uses the AStar heuristic based algorithm to
 * determine a path.
 * 
 * @author Kevin Glass
 */
public class PathFinder {
    /** The set of nodes that have been searched through */
    private final ArrayList<Node> f04;
    /** The set of nodes that we do not yet consider fully searched */
    private final SortedNodeList f14 = new SortedNodeList();

    /** The map being searched */
    private final Map f24;
    /** The maximum depth of search we're willing to accept before giving up */
    private final int f34;

    /** The complete set of nodes across the map */
    private final Node[][] f44;
    /** True if we allow diaganol movement */
    private final boolean f54;
    /** The heuristic we're applying to determine which nodes to search first */
    private final AStarHeuristic f64;

    /**
     * Create a path finder with the default heuristic - closest to target.
     * 
     * @param v0
     *            The map to be searched
     * @param v1
     *            The maximum depth we'll search before giving up
     * @param v2
     *            True if the search should try diaganol movement
     */
    public PathFinder(Map v0, int v1, boolean v2) {
        this(v0, v1, v2, new AStarHeuristic());
    }

    /**
     * Create a path finder
     * 
     * @param v6
     *            The heuristic used to determine the search order of the map
     * @param v3
     *            The map to be searched
     * @param v4
     *            The maximum depth we'll search before giving up
     * @param v5
     *            True if the search should try diaganol movement
     */
    public PathFinder(Map v3, int v4, boolean v5,
            AStarHeuristic v6) {
        this.f64 = v6;
        this.f24 = v3;
        this.f34 = v4;
        this.f54 = v5;

        f04 = new ArrayList<Node>();
        f44 = new Node[v3.m05()][v3.m15()];

        int v7;
		for (v7 = 0; v7 < v3.m05(); v7 = v7 + 1) {
            int v8;
			for (v8 = 0; v8 < v3.m15(); v8 = v8 + 1) {
                f44[v7][v8] = new Node(v7, v8);
            }
        }
    }

    /**
     * Find a path from the starting location provided (sx,sy) to the target
     * location (tx,ty) avoiding blockages and attempting to honour costs
     * provided by the tile map.
     * 
     * @param v9
     *            The entity that will be moving along the path. This provides a
     *            place to pass context information about the game entity doing
     *            the moving, e.g. can it fly? can it swim etc.
     * 
     * @param v10
     *            The x coordinate of the start location
     * @param v11
     *            The y coordinate of the start location
     * @param v12
     *            The x coordinate of the target location
     * @param v13
     *            Teh y coordinate of the target location
     * @return The path found from start to end, or null if no path can be
     *         found.
     */

    public Path m04(Actor v9, int v10, int v11, int v12, int v13) {
        // easy first check, if the destination is blocked, we can't get there

        if (!f24.m195(v9, v12, v13)) {
            return null;
        }

        // initial state for A*. The closed group is empty. Only the starting

        // tile is in the open list and it'e're already there
        f44[v10][v11].f24 = 0;
        f44[v10][v11].f54 = 0;
        f04.clear();
        f14.m14();
        f14.m24(f44[v10][v11]);

        f44[v12][v13].f34 = null;

        int v14;
		v14 = 0;
        while ((v14 < f34) && (f14.m44() != 0)) {
            // pull out the first node in our open list, this is determined to

            // be the most likely to be the next step based on our heuristic

            final Node v15;
			v15 = m14();
            if (v15 == f44[v12][v13]) {
                break;
            }

            m44(v15);
            m54(v15);

            // search through all the neighbours of the current node evaluating

            // them as next steps

            int v16;
			for (v16 = -1; v16 < 2; v16 = v16 + 1) {
                int v17;
				for (v17 = -1; v17 < 2; v17 = v17 + 1) {
                    // not a neighbour, its the current tile

                    if ((v16 == 0) && (v17 == 0)) {
                        continue;
                    }

                    // if we're not allowing diaganol movement then only

                    // one of x or y can be set

                    if (!f54) {
                        if ((v16 != 0) && (v17 != 0)) {
                            continue;
                        }
                    }

                    // determine the location of the neighbour and evaluate it

                    final int v18;
					v18 = v16 + v15.f04;
                    final int v19;
					v19 = v17 + v15.f14;

                    if (m84(v9, v10, v11, v18, v19)) {
                        // the cost to get to this node is cost the current plus
                        // the movement

                        // cost to reach this node. Note that the heursitic
                        // value is only used

                        // in the sorted open list

                        final float v20;
						v20 = v15.f24
								+ m94(v9, v15.f04, v15.f14, v18, v19);
                        final Node v21;
						v21 = f44[v18][v19];

                        // if the new cost we've determined for this node is
                        // lower than

                        // it has been previously makes sure the node hasn'e've
                        // determined that there might have been a better path
                        // to get to

                        // this node so it needs to be re-evaluated

                        if (v20 < v21.f24) {
                            if (m34(v21)) {
                                m44(v21);
                            }
                            if (m64(v21)) {
                                m74(v21);
                            }
                        }

                        // if the node hasn't already been processed and
                        // discarded then

                        // reset it's cost to our current cost and add it as a
                        // next possible

                        // step (i.e. to the open list)

                        if (!m34(v21) && !(m64(v21))) {
                            v21.f24 = v20;
                            v21.f44 = m104(v9, v18, v19, v12, v13);
                            v14 = Math.max(v14, v21.m04(v15));
                            m24(v21);
                        }
                    }
                }
            }
        }

        // since we'e've run out of search
        // there was no path. Just return null

        if (f44[v12][v13].f34 == null) {
            return null;
        }

        // At this point we've definitely found a path so we can uses the parent

        // references of the nodes to find out way from the target location back

        // to the start recording the nodes on the way.

        final Path v22;
		v22 = new Path();
        Node v23;
		v23 = f44[v12][v13];
        while (v23 != f44[v10][v11]) {
            v22.m53(v23.f04, v23.f14);
            v23 = v23.f34;
        }
        v22.m53(v10, v11);

        // thats it, we have our path

        return v22;
    }

    /**
     * Get the first element from the open list. This is the next one to be
     * searched.
     * 
     * @return The first element in the open list
     */
    protected Node m14() {
        return (Node) f14.m04();
    }

    /**
     * Add a node to the open list
     * 
     * @param v24
     *            The node to be added to the open list
     */
    protected void m24(Node v24) {
        f14.m24(v24);
    }

    /**
     * Check if a node is in the open list
     * 
     * @param v25
     *            The node to check for
     * @return True if the node given is in the open list
     */
    protected boolean m34(Node v25) {
        return f14.m54(v25);
    }

    /**
     * Remove a node from the open list
     * 
     * @param v26
     *            The node to remove from the open list
     */
    protected void m44(Node v26) {
        f14.m34(v26);
    }

    /**
     * Add a node to the closed list
     * 
     * @param v27
     *            The node to add to the closed list
     */
    protected void m54(Node v27) {
        f04.add(v27);
    }

    /**
     * Check if the node supplied is in the closed list
     * 
     * @param v28
     *            The node to search for
     * @return True if the node specified is in the closed list
     */
    protected boolean m64(Node v28) {
        return f04.contains(v28);
    }

    /**
     * Remove a node from the closed list
     * 
     * @param v29
     *            The node to remove from the closed list
     */
    protected void m74(Node v29) {
        f04.remove(v29);
    }

    /**
     * Check if a given location is valid for the supplied mover
     * 
     * @param v30
     *            The mover that would hold a given location
     * @param v31
     *            The starting x coordinate
     * @param v32
     *            The starting y coordinate
     * @param v33
     *            The x coordinate of the location to check
     * @param v34
     *            The y coordinate of the location to check
     * @return True if the location is valid for the given mover
     */
    protected boolean m84(Actor v30, int v31, int v32, int v33, int v34) {
        boolean v35;
		v35 = (v33 < 0) || (v34 < 0) || (v33 >= f24.m05())
				|| (v34 >= f24.m15());

        if ((!v35) && ((v31 != v33) || (v32 != v34))) {
            v35 = f24.m195(v30, v33, v34) == false;
        }

        return !v35;
    }

    /**
     * Get the cost to move through a given location
     * 
     * @param v36
     *            The entity that is being moved
     * @param v37
     *            The x coordinate of the tile whose cost is being determined
     * @param v38
     *            The y coordiante of the tile whose cost is being determined
     * @param v39
     *            The x coordinate of the target location
     * @param v40
     *            The y coordinate of the target location
     * @return The cost of movement through the given tile
     */
    public float m94(Actor v36, int v37, int v38, int v39, int v40) {
        return f24.m225(v36, v37, v38, v39, v40);
    }

    /**
     * Get the heuristic cost for the given location. This determines in which
     * order the locations are processed.
     * 
     * @param v41
     *            The entity that is being moved
     * @param v42
     *            The x coordinate of the tile whose cost is being determined
     * @param v43
     *            The y coordiante of the tile whose cost is being determined
     * @param v44
     *            The x coordinate of the target location
     * @param v45
     *            The y coordinate of the target location
     * @return The heuristic cost assigned to the tile
     */
    public float m104(Actor v41, int v42, int v43, int v44, int v45) {
        return f64.m06(f24, v41, v42, v43, v44, v45);
    }

    /**
     * A simple sorted list
     * 
     * @author kevin
     */
    private class SortedNodeList {
        /** The list of elements */
        private final ArrayList<Node> f04 = new ArrayList<Node>();

        /**
         * Retrieve the first element from the list
         * 
         * @return The first element from the list
         */
        public Object m04() {
            return f04.get(0);
        }

        /**
         * Empty the list
         */
        public void m14() {
            f04.clear();
        }

        /**
         * Add an element to the list - causes sorting
         * 
         * @param v46
         *            The element to add
         */
        public void m24(Node v46) {
            f04.add(v46);
            Collections.sort(f04);
        }

        /**
         * Remove an element from the list
         * 
         * @param v47
         *            The element to remove
         */
        public void m34(Object v47) {
            f04.remove(v47);
        }

        /**
         * Get the number of elements in the list
         * 
         * @return The number of element in the list
         */
        public int m44() {
            return f04.size();
        }

        /**
         * Check if an element is in the list
         * 
         * @param v48
         *            The element to search for
         * @return True if the element is in the list
         */
        public boolean m54(Object v48) {
            return f04.contains(v48);
        }
    }

    /**
     * A single node in the search graph
     */
    private class Node implements Comparable<Object> {
        /** The x coordinate of the node */
        private final int f04;
        /** The y coordinate of the node */
        private final int f14;
        /** The path cost for this node */
        private float f24;
        /** The parent of this node, how we reached it in the search */
        private Node f34;
        /** The heuristic cost of this node */
        private float f44;
        /** The search depth of this node */
        private int f54;

        /**
         * Create a new node
         * 
         * @param v49
         *            The x coordinate of the node
         * @param v50
         *            The y coordinate of the node
         */
        public Node(int v49, int v50) {
            this.f04 = v49;
            this.f14 = v50;
        }

        /**
         * Set the parent of this node
         * 
         * @param v51
         *            The parent node which lead us to this node
         * @return The depth we have no reached in searching
         */
        public int m04(Node v51) {
            f54 = v51.f54 + 1;
            this.f34 = v51;

            return f54;
        }

        /**
         * @see Comparable#compareTo(Object)
         */
        @Override
        public int compareTo(Object v52) {
            final Node v53;
			v53 = (Node) v52;

            final float v54;
			v54 = f44 + f24;
            final float v55;
			v55 = v53.f44 + v53.f24;

            if (v54 < v55) {
                return -1;
            } else if (v54 > v55) {
                return 1;
            } else {
                return 0;
            }
        }
    }
}
