package pacman.map;

import java.awt.Color;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;

import pacman.actors.Actor;
import pacman.actors.Ghost;
import pacman.actors.Player;
import pacman.game.GameObject;
import pacman.game.Item;
import pacman.util.Direction;

/**
 * Map class keeps track objects on the playing grid, helper methods to make
 * movement decisions, and export/import methods for the editor
 * 
 * @author Ramsey Kant
 */
public class Map {
    // Map parameters (width & height represent # of cells)
    private int f09;
    private int f19;
    public final int f29;
    public final int f39;
    public final int f49;
    public final double f59;

    // Instance vars
    private byte f69[][];
    private Item f79[][];
    private ArrayList<Actor> f89;
    private int f99;

    /**
     * Class constructor, inits a blank map based on a width, height, and cell
     * size Used in the editor
     * 
     * @param v0
     *            Width of the map
     * @param v1
     *            Height of the map
     * @param cs
     *            Size of individual cells in pixels
     */
    public Map(int v0, int v1, double v2) {
        // Set map parameters
        f09 = v0;
        f19 = v1;
        f59 = v2;
        f29 = (int) (32 * v2);
        f39 = (int) (12 * v2);
        f49 = (int) (10 * v2);
        f99 = 0;

        // Initialize collideMap, a 2D array that contains all static collidable
        // GameObjects
        // We use this for fast lookup during collision detection and AI
        // movement paths
        f69 = new byte[f09][f19];

        // Initialize itemMap, a 2D array that contains items (dots, powerups,
        // cherry) on the map
        f79 = new Item[f09][f19];

        // Create m_objects, an arraylist with all actorList
        f89 = new ArrayList<Actor>();
    }

    /**
     * Class Constructor that reads the map data from filename
     * 
     * @param v3
     *            The file name of the map to read contents from
     * @param cs
     *            Size of individual cells in pixels. This is something that
     *            should be deteremined by graphics, not the mapfile
     */
    public Map(String v3, double v4) {
        // Set the cell size
        f59 = v4;
        f29 = (int) (32 * v4);
        f39 = (int) (12 * v4);
        f49 = (int) (10 * v4);

        // Read contents of the map file
        m249(v3);
    }

    /**
     * The width of the map originally set in the constructor
     * 
     * @return The width of the map
     */
    public int m09() {
        return f09;
    }

    /**
     * The height of the map originally set in the constructor
     * 
     * @return The height of the map
     */
    public int m19() {
        return f19;
    }

    /**
     * Get the number of actorList on the map (the size of the actorList
     * ArrayList)
     * 
     * @return Number of actorList
     */
    public int m29() {
        return f89.size();
    }

    /**
     * Return the collidable map (a 2d array of bytes which correspond to the
     * collidable types defined in GameObject)
     * 
     * @return collidable map (collideMap)
     */
    public byte[][] m39() {
        return f69;
    }

    /**
     * Return the item map (a 2D array of Item objects)
     * 
     * @return item map (itemMap)
     */
    public Item[][] m49() {
        return f79;
    }

    /**
     * Return the number of dots remaining on the map. This is tracked by the
     * dotsRemaining local var (not a loop and count in itemMap)
     * 
     * @return dots remaining
     */
    public int m59() {
        return f99;
    }

    /**
     * Add a collidable (by type) to the collideMap
     * 
     * @param v5
     *            X coordinate
     * @param v6
     *            Y coordinate
     * @param v7
     *            Type of collidable
     * @return True if successful
     */
    public boolean m69(int v5, int v6, byte v7) {
        // Check bounds
        if (v5 < 0 || v6 < 0 || v5 >= f09 || v6 >= f19) {
            return false;
        }

        // Check if theres already something there
        if (f69[v5][v6] > 0) {
            return false;
        }

        // Add to the collideMap
        f69[v5][v6] = v7;
        return true;
    }

    /**
     * Put a new item to the item map
     * 
     * @param v8
     *            Item
     * @return True if successful
     */
    public boolean m79(Item v8) {
        if (v8 == null) {
            return false;
        }

        final int v9;
		v9 = v8.m320();
        final int v10;
		v10 = v8.m420();
        if (v9 < 0 || v10 < 0 || v9 >= f09 || v10 >= f19) {
            return false;
        }

        // Add to the itemMap
        if (v8.m020() == GameObject.f020) {
            f99++;
        }
        f79[v9][v10] = v8;
        return true;
    }

    /**
     * Put a new actor in the map (actorList ArrayList)
     * 
     * @param v11
     *            Actor
     * @return True if successful
     */
    public boolean m89(Actor v11) {
        if (v11 == null) {
            return false;
        }

        final int v12;
		v12 = v11.m320();
        final int v13;
		v13 = v11.m420();
        if (v12 < 0 || v13 < 0 || v12 >= f09 || v13 >= f19) {
            return false;
        }

        // Add to the array list
        f89.add(v11);
        return true;
    }

    /**
     * Return a value at (x,y) in the collision map
     * 
     * @param v14
     *            X Coordinate
     * @param v15
     *            Y Coordinate
     * @return Integer that represents the collision object
     */
    public byte m99(int v14, int v15) {
        // Check bounds
        if (v14 < 0 || v15 < 0 || v14 >= f09 || v15 >= f19) {
            return -1;
        }

        return f69[v14][v15];
    }

    /**
     * Return an item at coordinate (x,y) from within the item map (itemMap)
     * 
     * @param v16
     *            X Coordinate
     * @param v17
     *            Y Coordinate
     * @return Item the item that is found at (x,y)
     */
    public Item m109(int v16, int v17) {
        // Check bounds
        if (v16 < 0 || v17 < 0 || v16 >= f09 || v17 >= f19) {
            return null;
        }

        return f79[v16][v17];
    }

    /**
     * Return an actor at index in the actorList ArrayList
     * 
     * @param v18
     *            Index in actorList
     * @return Actor (null if non-existant)
     */
    public Actor m119(int v18) {
        Actor v19;
		v19 = null;
        try {
            v19 = f89.get(v18);
        } catch (final IndexOutOfBoundsException v20) {
            v20.printStackTrace();
        }
        return v19;
    }

    /**
     * Find and return the player object within the local actorList ArrayList
     * 
     * @return The player object. null if not found
     */
    public Player m129() {
        // Get from the object map
        for (final Actor v21 : f89) {
            if (v21.m020() == GameObject.f320) {
                return (Player) v21;
            }
        }

        return null;
    }

    /**
     * Return an actor at coordinate (x,y)
     * 
     * @param v22
     *            X Coordinate
     * @param v23
     *            Y Coordinate
     * @param v24
     *            If true, ignore a "Player" actor at (x,y)
     * @return Actor (null if an actor doesn't exist at the position)
     */
    public Actor m139(int v22, int v23, boolean v24) {
        // Check bounds
        if (v22 < 0 || v23 < 0 || v22 >= f09 || v23 >= f19) {
            return null;
        }

        // Get from the object map
        for (final Actor v25 : f89) {
            if (v24 && v25.m020() == GameObject.f320) {
                continue;
            }

            if (v25.m320() == v22 && v25.m420() == v23) {
                return v25;
            }
        }

        return null;
    }

    /**
     * Remove an actor from actorList based on index. Be careful when using
     * this! Just because an actor isn't in the map doesn't mean it's not
     * 'alive' This is primarily for the editor
     * 
     * @param v26
     *            Index of the actor
     */
    public void m149(int v26) {
        f89.remove(v26);
    }

    /**
     * Remove an item from the item array by coordinate (x, y)
     * 
     * @param v27
     *            X coordinate of the item
     * @param v28
     *            Y coordinate of the item
     */
    public void m159(int v27, int v28) {
        // Check bounds
        if (v27 < 0 || v28 < 0 || v27 >= f09 || v28 >= f19) {
            return;
        }

        if (f79[v27][v28].m020() == GameObject.f020) {
            f99--;
        }

        f79[v27][v28] = null;
    }

    /**
     * Remove everything at coordiante (x,y) Used by the editor only
     * 
     * @param v29
     *            X coordinate
     * @param v30
     *            Y coordinate
     * @return boolean True if something was removed, false if otherwise
     */
    public boolean m169(int v29, int v30) {
        boolean v31;
		v31 = false;

        // Check bounds
        if (v29 < 0 || v30 < 0 || v29 >= f09 || v30 >= f19) {
            return false;
        }

        // Remove any collidable
        if (f69[v29][v30] != 0) {
            f69[v29][v30] = 0;
            v31 = true;
        }

        // Remove any item
        if (f79[v29][v30] != null) {
            f79[v29][v30] = null;
            v31 = true;
        }

        int v32;
		v32 = 0;
		while (v32 < f89.size()) {
			Actor v33;
			v33 = f89.get(v32);
			if (v33.m320() == v29 && v33.m420() == v30) {
				f89.remove(v32);
				v33 = null;
				v32--;
				v31 = true;
			}
			v32++;
		}

        return v31;
    }

    /**
     * Find the distance (Manhattan) between two objects
     * 
     * @param v34
     *            GameObject at the initial position
     * @param v35
     *            GameObject at the end position
     * @return Distance (integer)
     */
    public int m179(GameObject v34, GameObject v35) {
        return (int) Math.sqrt(Math.pow(Math.abs(v34.m320() - v35.m320()), 2)
                + Math.pow(Math.abs(v34.m420() - v35.m420()), 2));
    }

    /**
     * Check if a coordinate is completely empty (void of actorList, items, and
     * collissions) Used by the editor
     * 
     * @param v36
     *            A x coordinate to move to
     * @param v37
     *            A y coordinate to move to
     * @return True if empty. False if otherwise
     */
    public boolean m189(int v36, int v37) {
        // Check bounds
        if (v36 < 0 || v37 < 0 || v36 >= f09 || v37 >= f19) {
            return false;
        }

        // Check if the Object is hitting something on the collideMap
        if (m99(v36, v37) != 0) {
            return false;
        }

        // Check if object is hitting something on the itemMap
        if (m109(v36, v37) != null) {
            return false;
        }

        // Actor collission
        if (m139(v36, v37, false) != null) {
            return false;
        }

        return true;
    }

    /**
     * Move attempt method. Changes the position the map of the game object if
     * there are no obstructions
     * 
     * @param v38
     *            The actor object trying to move
     * @param v39
     *            A x coordinate to move to
     * @param v40
     *            A y coordinate to move to
     * @return True if the move succeeded. False if otherwise
     */
    public boolean m199(Actor v38, int v39, int v40) {
        if (v38 == null) {
            return false;
        }

        // Check bounds
        if (!m219(v39, v40)) {
            return false;
        }

        // Check if the Object is hitting something on the collideMap
        if (m99(v39, v40) != 0) {
            return false;
        }

        // Allow the Actor to move
        return true;
    }

    public boolean m209(Actor v41, Direction v42) {
        int v43;
		v43 = v41.m320();
        int v44;
		v44 = v41.m420();
        switch (v42) {
            case f07:
                v44--;
                break;
            case f17:
                v43++;
                break;
            case f27:
                v44++;
                break;
            case f37:
                v43--;
                break;
            case f47:
                return true;
        }

        return m199(v41, v43, v44);
    }

    private boolean m219(int v45, int v46) {
        return v45 > 0 && v46 > 0 && v45 <= f09 && v46 <= f19;
    }

    /**
     * Get the cost of moving through the given tile. This can be used to make
     * certain areas more desirable. A simple and valid implementation of this
     * method would be to return 1 in all cases.
     * 
     * @param v47
     *            The mover that is trying to move across the tile
     * @param v48
     *            The x coordinate of the tile we're moving from
     * @param v49
     *            The y coordinate of the tile we're moving from
     * @param v50
     *            The x coordinate of the tile we're moving to
     * @param v51
     *            The y coordinate of the tile we're moving to
     * @return The relative cost of moving across the given tile
     */
    public float m229(Actor v47, int v48, int v49, int v50, int v51) {
        return 1;
    }

    /**
     * Write the contents of this map to a file in the correct format
     * 
     * @param v52
     *            File name of the map
     */
    public void m239(String v52) {
        FileOutputStream v53;
        DataOutputStream v54;

        try {
            v53 = new FileOutputStream(v52);
            v54 = new DataOutputStream(v53);

            // Write the map file magic
            v54.writeUTF("RKPACMAP");

            // Write map width & height
            v54.writeInt(f09);
            v54.writeInt(f19);

            int v55;
			v55 = 0;
			while (v55 < f09) {
				int v56;
				v56 = 0;
				while (v56 < f19) {
					v54.write(f69[v55][v56]);
					v56++;
				}
				v55++;
			}

            Item v57;
			v57 = null;
            int v58;
			v58 = 0;
			while (v58 < f09) {
				int v59;
				v59 = 0;
				while (v59 < f19) {
					v57 = f79[v58][v59];
					if (v57 == null) {
						v54.writeBoolean(false);
						continue;
					}
					v54.writeBoolean(true);
					v54.writeInt(v57.m020());
					v54.writeInt(v57.m320());
					v54.writeInt(v57.m420());
					v54.writeInt(v57.m120().getRGB());
					if (v57.m020() == GameObject.f720) {
						v54.writeInt(v57.m112());
						v54.writeInt(v57.m212());
					}
					v59++;
				}
				v58++;
			}

            // Write the number of actorList, then all actor data
            v54.writeInt(f89.size());
            for (final Actor v60 : f89) {
                v54.writeInt(v60.m020());
                v54.writeInt(v60.m320());
                v54.writeInt(v60.m420());
                v54.writeInt(v60.m120().getRGB());
                if (v60.m020() == GameObject.f420) {
                    v54.writeBoolean(((Ghost) v60).m24());
                }
            }

            v54.close();
            v53.close();
        } catch (final IOException v61) {
            System.out.println("Failed to write map file: " + v61.getMessage());
        }
    }

    /**
     * Read a file with map contents and set the properties in this map Called
     * by the constructor.
     * 
     * @param v62
     *            File name of the map
     */
    private void m249(String v62) {

        FileInputStream v63;
        DataInputStream v64;

        try {
            v63 = new FileInputStream(v62);
            v64 = new DataInputStream(v63);

            // Check for the magic
            if (!v64.readUTF().equals("RKPACMAP")) {
                System.out.println("Not a map file!");
                return;
            }

            // Read map width & height
            f09 = v64.readInt();
            f19 = v64.readInt();
            f99 = 0;

            // Initialize collideMap, a 2D array that contains all static
            // collidable GameObjects
            // We use this for fast lookup during collision detection and AI
            // movement paths
            f69 = new byte[f09][f19];

            // Initialize itemMap, a 2D array that contains items (dots,
            // powerups, cherry) on the map
            f79 = new Item[f09][f19];

            // Create m_objects, an arraylist with all actorList
            f89 = new ArrayList<Actor>();

            int v65;
			v65 = 0;
			while (v65 < f09) {
				int v66;
				v66 = 0;
				while (v66 < f19) {
					m69(v65, v66, v64.readByte());
					v66++;
				}
				v65++;
			}

            int v67;
			v67 = 0;
			while (v67 < f09) {
				int v68;
				v68 = 0;
				while (v68 < f19) {
					if (!v64.readBoolean()) {
						continue;
					}
					final int v69;
					v69 = v64.readInt();
					final int v70;
					v70 = v64.readInt();
					final int v71;
					v71 = v64.readInt();
					final Color v72;
					v72 = new Color(v64.readInt());
					m79(new Item(v69, v72, this, v70, v71));
					if (v69 == GameObject.f720) {
						final int v73;
						v73 = v64.readInt();
						final int v74;
						v74 = v64.readInt();
						f79[v70][v71].m012(v73, v74);
					}
					v68++;
				}
				v67++;
			}

            final int v75;
			v75 = v64.readInt();
            int v76;
			v76 = 0;
			while (v76 < v75) {
				final int v77;
				v77 = v64.readInt();
				final int v78;
				v78 = v64.readInt();
				final int v79;
				v79 = v64.readInt();
				final Color v80;
				v80 = new Color(v64.readInt());
				if (v77 == GameObject.f320) {
					m89(new Player(this, v78, v79));
				} else if (v77 == GameObject.f420) {
					final boolean v81;
					v81 = v64.readBoolean();
					m89(new Ghost(v80, this, v78, v79, v81));
				}
				v76++;
			}

            v64.close();
            v63.close();
        } catch (final IOException v82) {
            System.out.println("Failed to read map file: " + v82.getMessage());
        }
    }

}
